<?php

/*
Plugin Name: photostream-sync
Description: Synchronize your public iCloud photostreams to your WordPress installation. Import images, Import videos, create gellery posts, and more.
Author: Demitrious Kelly <apokalyptik@apokalyptik.com>
Version: 2.0.3
Author URI: http://blog.apokalyptik.com/
*/

require dirname( __FILE__ ) . '/class-photostream-client.php';

class Photostream {

	var $streams = null;
	var $streaminfo = array();
	var $processing_errors = array();
	var $dev = false;

	function __construct() {
		add_action( 'init', array( $this, 'init' ) );
		$this->processing_errors = array(
			'new' => array( 'validation' => array() ),
		);
		if ( defined( 'PHOTOSTREAM_SYNC_DEV' ) && PHOTOSTREAM_SYNC_DEV )
			$this->dev = true;
		if ( $this->dev ) {
			add_action( 'wp_ajax_ps_stream_info', array( $this, 'ajax_stream_info' ) );
			add_action( 'wp_ajax_ps_stream_data', array( $this, 'ajax_stream_data' ) );
			add_action( 'wp_ajax_ps_stream_process', array( $this, 'ajax_stream_process' ) );
		}
	}

	function ajax_stream_info() {
		die( '<pre>'.print_r( $this->get_stream_info( $_POST['stream'] ), true ).'</pre>' );
	}

	function ajax_stream_data() {
		$client = new Photostream_Client( $_POST['stream'] );
		$client->get();
		print( '<pre>'.print_r( $client, true ).'</pre>' );
		unset( $client );
		die();
	}

	function ajax_stream_process() {
		$start = microtime(1);
		$stream = $this->get_stream_info( $_POST['stream'] );
		$stream->processed_groups = array();
		$this->process_stream( $stream );
		die( sprintf("in %s seconds", (microtime(1) - $start) ) );
	}

	function init() {
		add_action( 'admin_menu', array( $this, 'add_admin_menu_page' ) );
	}

	function add_admin_menu_page() {
		load_plugin_textdomain( 'photostream' );
		$page = add_submenu_page(
			'plugins.php',
			esc_html( __( 'Photostream Import Settings' ) ),
			esc_html( __( 'Photostreams' ) ),
			'import',
			'photostream',
			array( $this, 'display_admin_menu_page' )
		);
		add_action( "load-$page", array( $this, 'admin_enqueue' ) );
	}

	function admin_enqueue() {
		$this->process_admin_page_post();
		wp_enqueue_script( 'photostream', plugin_dir_url( __FILE__ ) . 'static/photostream.js',                array( 'jquery' ), '1.0', true );
		wp_enqueue_script( 'qtip',        plugin_dir_url( __FILE__ ) . 'static/jquery.qtip-1.0.0-rc3.min.js',  array( 'jquery' ), '1.0', true );
		wp_enqueue_style(  'photostream', plugin_dir_url( __FILE__ ) . 'static/photostream.css',               array(),           '1.0'       );
	}

	function display_text_transform_options() {
		?><div class="transform_options">
		<a href="#" onClick="jQuery(this).parent().find('div:first').toggle('fast'); return false;">Show Options</a>
		<div>
		stream-&gt;title, stream-&gt;key<br/>
		owner-&gt;firstName, owner-&gt;lastName, owner-&gt;fullName<br/>
		image-&gt;caption, image-&gt;guid, image-&gt;date(<a href="http://us2.php.net/manual/en/function.date.php" target="_blank">format</a>)<br/>
		batch-&gt;guid, batch-&gt;date(<a href="http://us2.php.net/manual/en/function.date.php" target="_blank">format</a>)<br/>
		fetch-&gt;date(<a href="http://us2.php.net/manual/en/function.date.php" target="_blank">format</a>)
		</div>
		</div>
		<?php 
	}

	function display_admin_menu_page() {
		$streams = get_option( "photostream_streams" );
		if ( empty( $streams ) )
			$streams = array();
		?>

		<h2 class="nav-tab-wrapper">
			<?php if ( !empty( $streams ) ): ?>
				<a href="#" for="manage" class="nav-tab nav-tab-active"><?php esc_html_e( 'Manage Streams' ); ?></a>
			<?php endif; ?>
			<a href="#" for="add" class="nav-tab <?php echo empty( $streams ) ? 'nav-tab-active' : ''; ?>"><?php esc_html_e( 'Add A Stream' ); ?></a>
		</h2>

		<div id="ps-sync-add" class="ps-sync-tab" style="<?php echo empty( $streams ) ? '' : 'display:none;'; ?>">
			<?php $this->admin_ui_add(); ?>
		</div>

		<div id="ps-sync-manage" class="ps-sync-tab" style="<?php echo empty( $streams ) ? 'display:none;' : ''; ?>">
			<?php $this->admin_ui_manage(); ?>
		</div>

		<?php
	}

	function admin_ui_manage() {
		$streams = $this->get_streams();
		$number_of_streams = count( $streams );
		foreach( $streams as $idx => $key ) { $stream = $this->get_stream_info( $key ); ?>
			<div class="ps-stream-manage-box">
				<div class="ps-stream-manage-toggle">
					<a 
						style="<? echo ( $number_of_streams == 1 ? 'display: none;' : '' ); ?>" 
						href="#"><button class="button" onClick="return false;"><?php esc_html_e( 'Edit PhotoStream' ); ?></button></a>
					<?php esc_html_e( $stream->title ); ?>
				</div>
				<div class="ps-stream-manage-conent" style="<?php echo $number_of_streams == 1 ? '' : 'display:none;'; ?>">
					<form method="post">
					<input type="hidden" name="section" value="edit" />
					<?php wp_nonce_field( 'ps-sync-edit-' . $key, 'photostream[edit][' . esc_attr( $key ) . '][ps-nonce]' ); ?>
					<table class="widefat">
						<tbody>
							<tr
								help="<?php esc_attr_e( 'The key for this photostream. Click on it to view the stream in your browser' ); ?>">
								<td><?php esc_html_e( 'Photo Stream URL' ); ?></td>
								<td>
									<a href="https://www.icloud.com/photostream/#<?php esc_attr_e( $key ); ?>"
										taget="_blank"><?php esc_html_e( $key ); ?></a>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Enable, or disable whether importing should happen for this stream' ); ?>">
								<td><?php esc_html_e( 'Enable Importing' ); ?></td>
								<td>
									<input type="hidden" name="photostream[edit][<?php esc_attr_e( $key );?>][enabled]" value="0"/>
									<input type="checkbox" name="photostream[edit][<?php esc_attr_e( $key );?>][enabled]" value="1"
										<?php checked( true, $stream->enabled ); ?>/>
								</td>
							</tr>
							
							<tr
								help="<?php esc_attr_e( 'Choose the WordPress user that this plugin is acting as when importing this photostream.' ); ?>">
								<td><?php esc_html_e( 'WordPress User' ); ?></td>
								<td>
									<?php wp_dropdown_users( array( 
										'name' => 'photostream[edit][' . esc_attr( $key ) . '][user]',
										'selected' => $stream->user ) ); ?>
								</td>
							</tr>

							<tr
								help="<?php esc_attr_e( 'Password Protect all gallery posts made during import' ); ?>">
								<td><?php esc_html_e( 'Password Protection' ); ?></td>
								<td>
									<input
									type="text" name="photostream[edit][<?php esc_attr_e( $key );?>][password]"
									value="<?php esc_attr_e( $stream->password ); ?>"/>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Configure the post title for gallery posts. Click "show options" for formatting help.' ); ?>">
								<td><?php esc_html_e( 'Gallery Post Titles' ); ?></td>
								<td>
									<input type="text" name="photostream[edit][<?php esc_attr_e( $key );?>][post_title]"
										value="<?php esc_attr_e( $stream->post_title ); ?>"
										placeholder="stream-&gt;title - batch-&gt;date(Y-m-d)" />
									<br/>
									<?php $this->display_text_transform_options(); ?>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Configure how a caption should be included in the gallery posts when one is available' ); ?>">
								<td><?php esc_html_e( 'Gallery Post Caption Use' ); ?></td>
								<td>
									<select name="photostream[edit][<?php esc_attr_e( $key );?>][use_caption]">
										<option value=""><?php esc_html_e( 'Do not use caption' ); ?></option>
										<option value="above" <?php selected( 'above', $stream->use_caption ); ?>><?php esc_html_e( "Place the caption above the gallery" ); ?></option>
										<option value="below" <?php selected( 'below', $stream->use_caption ); ?>><?php esc_html_e( "Place the caption below the gallery" ); ?></option>
									</select>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Choose whether gallery posts should be published automatically or kept as drafts.' ); ?>">
								<td><?php esc_html_e( 'Gallery Publishing'); ?></td>
								<td>
									<select name="photostream[edit][<?php esc_attr_e( $key );?>][post_status]">
										<option value="publish" <?php selected( "publish", $stream->post_status ); ?>>Publish</option>
										<option value="draft" <?php selected( "draft", $stream->post_status ); ?>>Draft</option>
									</select>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Choose one or more tags for gallery posts.' ); ?>">
								<td><?php esc_html_e( 'Gallery Tags' ); ?></td>
								<td>
									<input 
										type="text" 
										name="photostream[edit][<?php esc_attr_e( $key );?>][tags]" 
										value="<?php esc_attr_e( implode( ', ', $stream->tags ) ); ?>" 
										placeholder="tag1, tag2, tag3"/>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Choose one or more categories for gallery posts.' ); ?>">
								<td><?php esc_html_e( 'Gallery Categories' ); ?></td>
								<td>	
									<ul><?php wp_category_checklist( 0, 0, $stream->cats ); ?></ul>
								</td>
							</tr>
							<tr
								help="<?php esc_attr_e( 'Choose a post type for gallery posts. This is an advanced setting for users writing themes around this plugin.' ); ?>">
								<td><?php esc_html_e( 'Gallery Post Type' ); ?></td>
								<td>
									<select name="photostream[edit][<?php esc_attr_e( $key );?>][post_type]">
									<?php foreach ( get_post_types() as $type ) {  ?>
										<option 
											<?php selected( $type, $stream->post_type ); ?> 
											value="<?php esc_attr_e( $type ); ?>"><?php esc_html_e( $type ); ?></option>
									<?php } ?>
									</select>
								</td>
							</tr>

							<tr
								help="<?php esc_attr_e( 'Choose a custom gallery shortcode for gallery posts. This is an advanced setting and is best left alone unless you have a specific reason to change it.' ); ?>">
								<td><?php esc_html_e( 'Gallery Shortcode' ); ?></td>
								<td>
									<input name="photostream[edit][<?php esc_attr_e( $key );?>][gallery_shortcode]" type="text" value="<?php esc_attr_e($stream->gallery_shortcode); ?>" placeholder="[gallery]"/>
								</td>
							</tr>

							<tr
								help="<?php esc_attr_e( 'Configure the filename to save imported images as. Click "show options" for formatting help.' ); ?>">
								<td><?php esc_html_e( 'Save Images As' ); ?></td>
								<td>
									<input 
										type="text" 
										name="photostream[edit][<?php esc_attr_e( $key );?>][rename_file]" 
										value="<?php esc_attr_e( $stream->rename_file ); ?>" 
										placeholder="stream->title - owner->fullName -- file->date(Y-m-d).jpg"/>
									<br/>
									<?php $this->display_text_transform_options(); ?>
								</td>
							</tr>
						</tbody>
					</table>
					<br/>	
					<input type="submit" class='ps-submit button-primary' value="save">
					<input type="hidden" class='ps-delete' name="photostream[edit][<?php esc_attr_e( $key );?>][delete]" value="" />
					<button class="button ps-delete" onClick="return false"><?php esc_html_e( 'delete' ); ?></button>
					<?php if ( $this->dev ): ?>
					<button class="button" onClick="var whence=this; jQuery.post(ajaxurl, {action:'ps_stream_info',stream:'<?php esc_attr_e($stream->key); ?>'}, function(response) { jQuery(whence).parent().parent().find('.debug').html(response); }); return false">info</button>
					<button class="button" onClick="var whence=this; jQuery.post(ajaxurl, {action:'ps_stream_data',stream:'<?php esc_attr_e($stream->key); ?>'}, function(response) { jQuery(whence).parent().parent().find('.debug').html(response); }); return false">data</button>
					<button class="button" onClick="var whence=this; jQuery.post(ajaxurl, {action:'ps_stream_process',stream:'<?php esc_attr_e($stream->key); ?>'}, function(response) { jQuery(whence).parent().parent().find('.debug').html(response); }); return false">process</button>
					<?php endif; ?>
					</form>
					<div class="debug"></div>
				</div>
			</div>
		<?php 
		}
	}

	function admin_ui_add() {
		?>
		<form method="post">
		<?php wp_nonce_field( 'ps-sync-add', 'ps-nonce' ); ?>
		<input type="hidden" name="section" value="add" />
		<table class="widefat">
			<tbody>
				<tr help="<?php esc_attr_e( 'The key for this photostream.' ); ?>">
					<td><?php esc_html_e( 'Photo Stream URL' ); ?></td>
					<td><input type="text" name="photostream[add][stream]" id="add_photostream_stream" placeholder="<?php esc_attr_e('Stream to add'); ?>"/></td>
				</tr>
				<tr help="<?php esc_attr_e( 'Enable, or disable whether importing should happen for this stream' ); ?>">
					<td><?php esc_html_e( 'Enable Importing' ); ?></td>
					<td>
						<input type="hidden" name="photostream[add][enabled]" value="0"/>
						<input type="checkbox" name="photostream[add][enabled]" value="1" checked/>
					</td>
				</tr>
				<tr help="<?php esc_attr_e( 'Choose the WordPress user that this plugin is acting as when importing this photostream.' ); ?>">
					<td><?php esc_html_e( 'WordPress User' ); ?></td>
					<td><?php wp_dropdown_users( array( 'name' => 'photostream[add][user]' ) ); ?></td>
				</tr>
				<tr help="<?php esc_attr_e( 'Password Protect all gallery posts made during import' ); ?>">
					<td><?php esc_html_e( 'Password Protection' ); ?></td>
					<td><input name="photostream[add][password]" type="text"/></td>
				</tr>
				<tr help="<?php esc_attr_e( 'Configure the post title for gallery posts. Click "show options" for formatting help.' ); ?>">
					<td><?php esc_html_e( 'Gallery Post Titles' ); ?></td>
					<td>
						<input type="text" name="photostream[add][post_title]" value=""
							placeholder="stream-&gt;title - batch-&gt;date(Y-m-d)" />
						<br/>
						<?php $this->display_text_transform_options(); ?>
					</td>
				</tr>
				<tr
					help="<?php esc_attr_e( 'Configure how a caption should be included in the gallery posts when one is available' ); ?>">
					<td><?php esc_html_e( 'Gallery Post Caption Use' ); ?></td>
					<td>
						<select name="photostream[add][use_caption]">
							<option value=""><?php esc_html_e( 'Do not use caption' ); ?></option>
							<option value="above"><?php esc_html_e( "Place the caption above the gallery" ); ?></option>
							<option value="below"><?php esc_html_e( "Place the caption below the gallery" ); ?></option>
						</select>
					</td>
				</tr>
				<tr help="<?php esc_attr_e( 'Choose whether gallery posts should be published automatically or kept as drafts.' ); ?>">
					<td><?php esc_html_e( 'Gallery Publishing' ); ?></td>
					<td>
						<select name="photostream[add][post_status]">
							<option value="publish">Publish</option>
							<option value="draft">Draft</option>
						</select>
					</td>
				</tr>
				<tr help="<?php esc_attr_e( 'Choose one or more tags for gallery posts.' ); ?>">
					<td><?php esc_html_e( 'Gallery Tags' ); ?></td>
					<td><input type="text" name="photostream[add][tags]" id="add_photostream_tags" placeholder="tag1, tag2, tag3"/></td>
				</tr>
				<tr help="<?php esc_attr_e( 'Choose one or more categories for gallery posts.' ); ?>">
					<td><?php esc_html_e( 'Gallery Categories' ); ?></td>
					<td><ul style="margin-left: 2em;"><?php wp_category_checklist( array( 'name' => 'photostream[add][categories]' ) ); ?></ul></td>
				</tr>
				<tr help="<?php esc_attr_e( 'Choose a post type for gallery posts. This is an advanced setting for users writing themes around this plugin.' ); ?>">
					<td><?php esc_html_e( 'Gallery Post Type' ); ?></td>
					<td>
						<select name="photostream[add][post_type]"><?php foreach ( get_post_types() as $type ) { ?>
						<option value="<?php esc_attr_e( $type ); ?>"><?php esc_html_e( $type ); ?></option>
						<?php } ?></select>
					</td>
				</tr>
				<tr
					help="<?php esc_attr_e( 'Choose a custom gallery shortcode for gallery posts. This is an advanced setting and is   best left alone unless you have a specific reason to change it.' ); ?>">
					<td><?php esc_html_e( 'Gallery Shortcode' ); ?></td>
					<td>
						<input name="photostream[add][gallery_shortcode]" type="text" placeholder="[gallery]"/>
					</td>
				</tr>
				<tr help="<?php esc_attr_e( 'Configure the filename to save imported images as. Click "show options" for formatting help.' ); ?>">
					<td><?php esc_html_e( 'Save Images As' ); ?></td>
					<td>
						<input 
							type="text" 
							name="photostream[add][rename_file]" 
							value="<?php esc_attr_e( $stream->rename_file ); ?>" 
							placeholder="stream->title - owner->fullName -- file->date(Y-m-d).jpg"/>
						<br/>
						<?php $this->display_text_transform_options(); ?>
					</td>	
				</tr>
			</tbody>
		</table>
		<br/>
		<input type="submit" class='button-primary' value="<?php esc_attr_e( 'save' ); ?>">
		</form>
		<?php
	}
	
	function process_admin_add_post() {
		if ( !wp_verify_nonce( $_POST['ps-nonce'], 'ps-sync-add' ) )
			return;
		$new_stream = stripslashes_deep( $_POST['photostream']['add'] );
		$key = $this->parse_user_input_photostream_key( $_POST['photostream']['add']['stream'] );
		$new_stream['key'] = $key;
		if ( !empty( $new_stream['tags'] ) )
			$new_stream['tags'] = preg_split( '/ *, */', $new_stream['tags'] );
		else
			$new_stream['tags'] = array();
		$new_stream['cats'] = empty( $_POST['post_category'] ) ? array() : $_POST['post_category'];
		$client = new Photostream_Client( $key );
		$new_stream['title'] = $client->title(); 
		if ( empty( $new_stream['title'] ) )
			$new_stream['title'] = $key;
		$this->add_stream( $new_stream );
	}

	function process_admin_edit_post() {
		foreach( $_POST['photostream']['edit'] as $key => $data ) {
			$data = stripslashes_deep( $data );
			if ( !wp_verify_nonce( $_POST['photostream']['edit'][$key]['ps-nonce'], 'ps-sync-edit-' . $key ) )
				continue;
			unset( $data['ps-nonce'] );
			if ( !empty( $data['delete'] ) ) {
				$this->remove_stream( $key );
				$client = new Photostream_Client( $key );
				$client->erase();
				continue;
			} else {
				unset( $data['delete'] );
				$stream = $this->get_stream_info( $key );
				if ( !empty( $data['tags'] ) )
					$data['tags'] = preg_split( '/ *, */', $data['tags'] );
				else
					$data['tags'] = array();
				foreach( $data as $option => $val )
					$stream->$option = $val;
				if ( !empty( $_POST['post_category'] ) )
					$stream->cats = $_POST['post_category'];
				else
					$stream->cats = array();
				$this->add_stream( $stream );
			}
		}
	}

	function process_admin_page_post() {
		if ( empty( $_POST ) || ( defined( 'DOING_AJAX' ) && DOING_AJAX ) )
			return;
		switch( $_POST['section'] ) {
			case 'general': $this->process_admin_general_post(); break;
			case 'add':     $this->process_admin_add_post();     break;
			case 'edit':    $this->process_admin_edit_post();    break;
		}
		wp_redirect( $_SERVER['REQUEST_URI'] );
	}

	function process_stream_group( &$stream, $group_key, &$images ) {
		if ( empty( $stream->processed_groups[$group_key] ) )
			$stream->processed_groups[$group_key] = array();
		$group_post_id = 0;
		$group_posts = get_posts( array( 
			'meta_key' => 'ps_batch_guid',
			'meta_value' => $group_key,
			'post_parent' => 0,
			'post_status' => array( 'draft', 'publish' ),
			'post_type' => $stream->post_type
		) );
		if ( empty( $group_posts ) ) {
			$post = array(
				'post_author' => $stream->user,
				'post_content' => '[gallery]',
				'post_date_gmt' => gmdate( 'Y-m-d H:i:s', strtotime( $images[0]->batchDateCreated ) ) ,
				'post_title' => 'stream->title - batch->date(Y-m-d)',
				'post_status' => 'draft',
				'post_type' => $stream->post_type,
				'tags_input' => empty( $stream->tags ) ? '' : implode( ', ', $stream->tags ),
			);
			if ( !empty( $stream->gallery_shortcode ) )
				$post['post_content'] = $stream->gallery_shortcode;
			if ( !empty( $images[0]->caption ) && $stream->use_caption ) {
				if ( $stream->use_caption == 'above' )
					$post['post_content'] = $images[0]->caption . "\n" . $post['post_content'];
				else
					$post['post_content'] = $post['post_content']. "\n" . $images[0]->caption;
			}
			if ( !empty( $stream->post_title ) )
				$post['post_title'] = $stream->post_title;
			$post['post_title'] = $this->apply_stream_centric_text_transforms( $post['post_title'], $stream );
			$post['post_title'] = $this->apply_image_centric_text_transforms( $post['post_title'], $images[0] );
			$post['post_title'] = $this->apply_generic_text_transforms( $post['post_title'] );
			if ( !empty( $stream->password ) )
				$post['post_password'] = $stream->password;
			$group_post_id = wp_insert_post( $post );
			if ( empty( $group_post_id ) )
				return false; // what?
			update_post_meta( $group_post_id, 'ps_stream', $stream->key );
			update_post_meta( $group_post_id, 'ps_batch_guid', $group_key );
			$group_post = get_post( $group_post_id, ARRAY_A );
			set_post_format( $group_post_id, 'gallery' );
		} else {
			$group_post_id = $group_posts[0]->ID;
			$group_post = get_post( $group_post_id, ARRAY_A );
		}

		if ( !empty( $stream->cats ) )
			wp_set_post_terms( $group_post_id, $stream->cats, 'category' );
		else
			wp_set_post_terms( $group_post_id, '', 'category' );

		if ( !empty( $stream->tags ) )
			wp_set_post_terms( $group_post_id, $stream->tags, 'post_tag' );
		else
			wp_set_post_terms( $group_post_id, '', 'post_tag' );

		$ready = true;
		foreach( $images as $image ) {
			if ( empty( $image ) )
				continue;
			if ( !empty( $stream->processed_groups[$group_key][$image->photoGuid] ) )
				continue;
			$processed = $this->process_group_photo( $stream, $group_post_id, $group_key, $image );
			if ( !$processed )
				$ready = false;
			if ( $processed['type'] == 'video/mp4' ) {
				$group_post['post_content'] = $group_post['post_content'] . sprintf( '[video mp4="%s"]', $processed['url'] );
				wp_update_post($group_post);
			}
		}
		if ( !$ready )
			return;
		$group_post = get_post( $group_post_id, ARRAY_A );
		$group_post['post_status'] = $stream->post_status;
		wp_update_post( $group_post );
	}

	function process_group_photo( &$stream, $group_post_id, $group_key, $image ) {
		$photo_post = get_posts( array( 
			'meta_key' => 'ps_photo_guid',
			'meta_value' => $image->photoGuid,
			'post_parent' => $group_post_id,
			'post_status' => null,
			'post_type' => 'attachment'
		) );
		if ( !empty( $photo_post ) )
			return true;

		if ( empty( $image ) )
			return false;
		if ( empty( $image->derivatives ) )
			return false;
		$derivatives = get_object_vars( $image->derivatives );
		$width = max( array_keys( $derivatives ) );
		
		if ( $stream->rename_file )
			$path = $stream->rename_file;
		else
			$path = 'stream->title - owner->fullName -- image->date(Y-m-d).jpg';
		$path = $this->apply_stream_centric_text_transforms( $path, $stream );
		$path = $this->apply_image_centric_text_transforms( $path, $image );
		$path = $this->apply_generic_text_transforms( $path );
		$path = preg_replace( '/\.(jpe?g|mp4)$/i', '', $path );
		if ( isset( $image->mediaAssetType ) && $image->mediaAssetType == 'video' )
			$path = "$path.mp4";
		else
			$path = "$path.jpg";
		if ( empty( $path ) )
			return false;
		$mime = wp_check_filetype( $path );
		if ( empty( $mime ) || empty( $mime['type'] ) )
			return false;
		if ( isset( $image->mediaAssetType ) && $image->mediaAssetType == 'video' ) {
			foreach( array( '720p', '360p' ) as $resolution ) {
				foreach( $derivatives[$resolution]->_wp_ps_url as $url ) {
					$bits = wp_remote_retrieve_body( $response = wp_remote_get( $url ) );
					if ( $bits )
						break;
				}
				if ( $bits )
					break;
			}
		} else {
			foreach( $derivatives[$width]->_wp_ps_url as $url ) {
				$bits = wp_remote_retrieve_body( $response = wp_remote_get( $url ) );
				if ( $bits )
					break;
			}
		}
		if ( empty( $bits ) )
			return false;
		if ( $bits === 'Unauthorized' )
			return false;
		$upload = wp_upload_bits( $path, null, $bits );
		if ( ! empty($upload['error']) )
			return false;
		$attachment = array(
			'post_title' => $path,
			'post_content' => '',
			'post_type' => 'attachment',
			'post_parent' => $group_post_id,
			'post_mime_type' => $mime['type'],
			'guid' => $upload[ 'url' ]
		);

		$exif = false;
		if ( !isset( $image->mediaAssetType ) || $image->mediaAssetType != 'video' ) {
			if ( function_exists( 'exif_read_data' ) )
				$exif = exif_read_data( $upload[ 'file' ] );
		}
		$id = wp_insert_attachment( $attachment, $upload[ 'file' ], $post_id );
		wp_update_attachment_metadata( $id, wp_generate_attachment_metadata( $id, $upload['file'] ) );
		update_post_meta( $id, 'ps_stream', $stream->key );
		update_post_meta( $id, 'ps_batch_guid', $group_post_id );
		update_post_meta( $id, 'ps_photo_guid', $image->photoGuid );
		update_post_meta( $id, 'ps_image_data', $image );
		if ( $exif && is_array( $exif ) ) { 
			foreach( $exif as $exif_idx => $exif_val ) {
				update_post_meta( $id, 'ps_exif_' . $exif_idx, $exif_val );
			}
		}
		$struct = array(
			'id'   => strval( $id ),
			'file' => $upload['file'],
			'url'  => $upload[ 'url' ],
			'type' => $mime['type']
		);
		$stream->processed_groups[$group_key][$image->photoGuid] = time();
		return apply_filters( 'wp_handle_upload', $struct, 'upload' );
	}

	function process_stream( $stream ) {
		$stream->last_processed = time();
		$stream->last_completed = time();
		if ( empty( $stream->processed_groups ) )
			$stream->processed_groups = array();
		$client = new Photostream_Client( $stream->key );
		if ( !$client->get() ) {
			unset( $client );
			return false;
		}
		$groups = $client->groups();
		foreach( array_reverse( array_keys( $groups ) ) as $group_key ) {
			$images = $groups[$group_key];
			$this->process_stream_group( $stream, $group_key, $images ); 
		}
		if ( !empty( $stream->processed_groups ) ) {
			foreach( $stream->processed_groups as $key => $group ) {
				if ( !empty( $groups[$key] ) )
					continue;
				unset( $stream->processed_groups[$key] );
			}
		}
		$this->add_stream( $stream );
		unset( $client );
	}

	function do_scheduled_work() {
		if ( !function_exists('wp_generate_attachment_metadata') )
			include ABSPATH . 'wp-admin/includes/image.php';
		if ( !function_exists('wp_read_video_metadata') )
			include ABSPATH . 'wp-admin/includes/media.php';
		$streams = get_option( "photostream_streams" );
		if ( empty( $streams ) )
			return;
		foreach( $streams as $key ) {
			$stream = $this->get_stream_info( $key );
			if ( !$stream->enabled )
				continue;
			$process = false;
			if ( empty( $stream->last_processed ) )
				$process = true;
			else if ( empty( $stream->last_completed ) )
				$process = true;
			else if ( $stream->last_completed < $stream->last_processed )
				$process = true;
			else if ( ( time() - $stream->last_completed ) > 1800 )
				$process = true;
			if ( !$process )
				continue;
			$this->process_stream( $stream );
		}
	}

	function apply_stream_centric_text_transforms( $text, $stream ) {
		if ( empty( $text ) || empty( $stream ) )
			return $text;
		$text = str_replace( 'stream->title', $stream->title, $text );
		$text = str_replace( 'stream->key', $stream->key, $text );
		return $text;
	}

	function apply_image_centric_text_transforms( $text, $image ) {
		if ( empty( $text ) || empty( $image ) )
			return $text;
		// Owner replacements
		$text = str_replace( 'owner->firstName', $image->contributorFirstName, $text );
		$text = str_replace( 'owner->lastName', $image->contributorLastName, $text );
		$text = str_replace( 'owner->fullName', $image->contributorFullName, $text );

		// Image Replacements
		$text = str_replace( 'image->', 'file->', $text ); // Maintain backwards compatibility
		$text = str_replace( 'video->', 'file->', $text ); // Just in case someone tries this...
		$text = str_replace( 'file->caption', $image->caption, $text );
		$text = str_replace( 'file->guid', $image->photoGuid, $text );
		if ( preg_match_all( '/file->date\(([^)]+)\)/', $text, $matches ) ) {
			$datetime = strtotime( $image->dateCreated );
			foreach( array_keys( $matches[0] ) as $m )
				$text = str_replace( $matches[0][$m], date( $matches[1][$m], $datetime ), $text );
		}

		// Batch Replacements
		$text = str_replace( 'batch->guid', $image->batchGuid, $text );
		if ( preg_match_all( '/batch->date\(([^)]+)\)/', $text, $matches ) ) {
			$datetime = strtotime( $image->batchDateCreated );
			foreach( array_keys( $matches[0] ) as $m )
				$text = str_replace( $matches[0][$m], date( $matches[1][$m], $datetime ), $text );
		}
		return $text;
	}

	function apply_generic_text_transforms( $text ) {
		if ( empty( $text ) )
			return $text;
		// Current DateTime Replacements
		if ( preg_match_all( '/fetch->date\(([^)]+)\)/', $text, $matches ) ) {
			foreach( array_keys( $matches[0] ) as $m )
				$text = str_replace( $matches[0][$m], date( $matches[1][$m] ), $text );
		}
		return $text;
	}

	function stream_option_key( $stream ) {
		return sprintf( "photostream_i_%s", preg_replace( '/[^0-9a-z]/i', '-', $stream ) );
	}
	
	function clear_thread_stream_cache() {
		$this->streams = null;
		$this->streaminfo = array();
	}

	function get_streams() {
		if ( null !== $this->streams )
			return $this->streams;
		$this->streams = get_option( "photostream_streams" );
		if ( empty( $this->streams ) )
			$this->streams = array();
		return $this->streams;
	}

	function add_stream( $stream ) {
		if ( is_array( $stream ) )
			$stream = (object)$stream;
		if ( !is_object( $stream ) )
			return false;
		$stream->_version = 1;
		update_option( $this->stream_option_key( $stream->key ), $stream );
		$list = $this->get_streams();
		$list[] = $stream->key;
		$list = array_filter( array_unique( $list ) );
		update_option( 'photostream_streams', $list );
		$this->clear_thread_stream_cache();
	}

	function remove_stream( $stream_key ) {
		delete_option( $this->stream_option_key( $stream_key ) );
		$list = preg_grep( 
			'/^' . preg_quote( $stream_key ).'$/',
			$this->get_streams(),
			PREG_GREP_INVERT
		);
		$list = array_filter( array_unique( $list ) );
		update_option( 'photostream_streams', $list );
		$this->clear_thread_stream_cache();
	}

	function get_stream_info( $stream ) {
		$key = $this->stream_option_key( $stream );
		if ( array_key_exists( $key, $this->streaminfo ) )
			return $this->streaminfo[$key];
		$this->streaminfo[$key] = get_option( $key );
		if ( empty( $this->streaminfo[$key] ) ) {
			$this->streaminfo[$key] = (object)array(
				'shard' => $stream{1},
			);
		}
		return $this->streaminfo[$key];
	}

	function parse_user_input_photostream_key( $key ) {
		if ( 0 === strpos( $key, '#' ) )
			$key = substr( $key, 1 );
		else if ( false !== strpos( $key, '//' ) || false !==  strpos( $key, '#' ) ) 
			$key = parse_url( $key, PHP_URL_FRAGMENT );
		if ( !preg_match( '/^[0-9a-zA-Z]+$/', $key ) )
			return false;
		return $key;
	}

	function add_processing_error( $section, $subsection, $severity, $message ) {
		$this->processing_errors[$section][$subsection][] = (object)array( 
			'severity' => $severity, 
			'message' => $message 
		);
	}

	function validate_photostream_key( $key ) {
		$rval = (object)array(
			'shard' => 1,
			'last_fetch_time' => 0,
			'debug' => '',
		);
		$clean_key = $this->parse_user_input_photostream_key( $key );
		if ( empty( $clean_key ) ) {
			$this->add_processing_error( 
				'new', 
				'validation', 
				0, 
				sprintf( __('%s does not look like a valid Photostream key.'), $key )
			);
			return false;
		}
	}

}

function photostream_cron() {
	$GLOBALS['__photostream']->do_scheduled_work();
}
add_action( 'photostream_hourly_cron', 'photostream_cron' );

function photostream_cron_add_quarter_hourly( $schedules ) {
	$schedules['quarter_hourly'] = array(
		'interval' => 900,
		'display' => __( 'Every 15 minutes' ),
	);
	return $schedules;
}
add_filter( 'cron_schedules', 'photostream_cron_add_quarter_hourly' );

function photostream_activation() {
	if ( !wp_next_scheduled( 'photostream_hourly_cron' ) )
		wp_schedule_event( time(), 'quarter_hourly', 'photostream_hourly_cron' );
}
register_activation_hook( __FILE__, 'photostream_activation' );

function photostream_deactivation() {
	if ( wp_next_scheduled( 'photostream_hourly_cron' ) ) 
		wp_clear_scheduled_hook('photostream_hourly_cron');
}
register_deactivation_hook( __FILE__, 'photostream_deactivation' );

$GLOBALS['__photostream'] = new Photostream();
